"""Flux-local diff action."""

import difflib
import logging
import pathlib

from flux_local import git_repo

from . import build

_LOGGER = logging.getLogger(__name__)


class DiffAction:
    """Flux-local diff action."""

    async def run(  # type: ignore[no-untyped-def]
        self,
        path: pathlib.Path,
        enable_helm: bool,
        skip_crds: bool,
        **kwargs,  # pylint: disable=unused-argument
    ) -> None:
        """Async Action implementation."""
        repo = git_repo.git_repo(path)

        content1 = []
        with git_repo.create_worktree(repo) as worktree_dir:
            async for content in build.build(worktree_dir, enable_helm, skip_crds):
                content1.append(content)

        content2 = []
        async for content in build.build(
            git_repo.repo_root(repo), enable_helm, skip_crds
        ):
            content2.append(content)

        diff_text = difflib.unified_diff(
            "".join(content1).split("\n"), "".join(content2).split("\n")
        )
        print("\n".join(diff_text))



"""Library for issuing commands using asyncio and returning the result."""

import asyncio
import logging
import shlex
import subprocess
from dataclasses import dataclass
from pathlib import Path

_LOGGER = logging.getLogger(__name__)


# No public API
__all__: list[str] = []


class CommandException(Exception):
    """Error while running a command."""


@dataclass
class Command:
    """An instance of a command to run."""

    cmd: list[str]
    """Array of command line arguments."""

    cwd: Path | None = None
    """Current working directory."""

    @property
    def string(self) -> str:
        """Render the command as a single string."""
        return " ".join([shlex.quote(arg) for arg in self.cmd])

    def __str__(self) -> str:
        """Render as a debug string."""
        return f"({self.cwd}) {self.string}"


async def run_piped(cmds: list[Command]) -> str:
    """Run a set of commands, piped together, returning stdout of last."""
    stdin = None
    out = None
    for cmd in cmds:
        _LOGGER.debug("Running command: %s", cmd)
        proc = await asyncio.create_subprocess_shell(
            cmd.string,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=cmd.cwd,
        )
        out, err = await proc.communicate(stdin)
        if proc.returncode:
            if out:
                _LOGGER.error(out.decode("utf-8"))
            if err:
                _LOGGER.error(err.decode("utf-8"))
            raise CommandException(
                f"Subprocess '{cmd}' failed with return code {proc.returncode}: "
                + out.decode("utf-8")
            )
        stdin = out
    return out.decode("utf-8") if out else ""


async def run(cmd: Command) -> str:
    """Run the specified command and return stdout."""
    return await run_piped([cmd])



"""Library for operating on a local repo and building Manifests.

This will build a `manifest.Manifest` from a cluster repo. This follows the
pattern of building kustomizations, then reading helm releases (though it will
not evaluate the templates). The resulting `Manifest` contains all the necessary
information to do basic checks on objects in the cluster (e.g. run templates
from unit tests).

Example usage:

```python
from flux_local import repo

manifest = await repo.build_manifest()
for cluster in manifest:
    print(f"Found cluster: {cluster.path}")
    for kustomization in cluster.kustomizations:
        print(f"Found kustomization: {kustomization.path}")
        for release in kustomization.helm_releases:
            print(f"Found helm release: {release.release_name}")
```

"""

import contextlib
import logging
import os
import tempfile
from collections.abc import Callable
from functools import cache
from pathlib import Path
from typing import Any, Generator

import git

from . import kustomize
from .manifest import (
    CLUSTER_KUSTOMIZE_DOMAIN,
    KUSTOMIZE_DOMAIN,
    Cluster,
    HelmRelease,
    HelmRepository,
    Kustomization,
    Manifest,
)

__all__ = [
    "repo_root",
    "build_manifest",
]

_LOGGER = logging.getLogger(__name__)

CLUSTER_KUSTOMIZE_NAME = "flux-system"
CLUSTER_KUSTOMIZE_KIND = "Kustomization"
KUSTOMIZE_KIND = "Kustomization"
HELM_REPO_KIND = "HelmRepository"
HELM_RELEASE_KIND = "HelmRelease"


def git_repo(path: Path | None = None) -> git.repo.Repo:
    """Return the local github repo path."""
    if path is None:
        return git.repo.Repo(os.getcwd(), search_parent_directories=True)
    _LOGGER.debug("Creating git repo: %s", path)
    return git.repo.Repo(str(path), search_parent_directories=True)


@cache
def repo_root(repo: git.repo.Repo | None = None) -> Path:
    """Return the local github repo path."""
    if repo is None:
        repo = git_repo()
    return Path(repo.git.rev_parse("--show-toplevel"))


def domain_filter(version: str) -> Callable[[dict[str, Any]], bool]:
    """Return a yaml doc filter for specified resource version."""

    def func(doc: dict[str, Any]) -> bool:
        if api_version := doc.get("apiVersion"):
            if api_version.startswith(version):
                return True
        return False

    return func


CLUSTER_KUSTOMIZE_DOMAIN_FILTER = domain_filter(CLUSTER_KUSTOMIZE_DOMAIN)
KUSTOMIZE_DOMAIN_FILTER = domain_filter(KUSTOMIZE_DOMAIN)


async def get_clusters(path: Path) -> list[Cluster]:
    """Load Cluster objects from the specified path."""
    cmd = kustomize.grep(f"kind={CLUSTER_KUSTOMIZE_KIND}", path).grep(
        f"metadata.name={CLUSTER_KUSTOMIZE_NAME}"
    )
    docs = await cmd.objects()
    return [
        Cluster.from_doc(doc) for doc in docs if CLUSTER_KUSTOMIZE_DOMAIN_FILTER(doc)
    ]


async def get_kustomizations(path: Path) -> list[Kustomization]:
    """Load Kustomization objects from the specified path."""
    cmd = kustomize.grep(f"kind={KUSTOMIZE_KIND}", path).grep(
        f"metadata.name={CLUSTER_KUSTOMIZE_NAME}",
        invert=True,
    )
    docs = await cmd.objects()
    return [Kustomization.from_doc(doc) for doc in docs if KUSTOMIZE_DOMAIN_FILTER(doc)]


async def build_manifest(path: Path | None = None) -> Manifest:
    """Build a Manifest object from the local cluster.

    This will locate all Kustomizations that represent clusters, then find all
    the Kustomizations within that cluster, as well as all relevant Helm
    resources.
    """
    root = repo_root(git_repo(path))

    clusters = await get_clusters(path or root)
    for cluster in clusters:
        _LOGGER.debug("Processing cluster: %s", cluster.path)
        cluster.kustomizations = await get_kustomizations(
            root / cluster.path.lstrip("./")
        )
        for kustomization in cluster.kustomizations:
            _LOGGER.debug("Processing kustomization: %s", kustomization.path)
            cmd = kustomize.build(root / kustomization.path)
            kustomization.helm_repos = [
                HelmRepository.from_doc(doc)
                for doc in await cmd.grep(f"kind=^{HELM_REPO_KIND}$").objects()
            ]
            kustomization.helm_releases = [
                HelmRelease.from_doc(doc)
                for doc in await cmd.grep(f"kind=^{HELM_RELEASE_KIND}$").objects()
            ]
    return Manifest(clusters=clusters)


@contextlib.contextmanager
def create_worktree(repo: git.repo.Repo) -> Generator[Path, None, None]:
    """Create a ContextManager for a new git worktree in the current repo.

    This is used to get a fork of the current repo without any local changes
    in order to produce a diff.
    """
    orig = os.getcwd()
    with tempfile.TemporaryDirectory() as tmp_dir:
        _LOGGER.debug("Creating worktree in %s", tmp_dir)
        repo.git.worktree("add", str(tmp_dir))
        os.chdir(tmp_dir)
        yield Path(tmp_dir)
    _LOGGER.debug("Restoring to %s", orig)
    # The temp directory should now be removed and this prunes the worktree
    repo.git.worktree("prune")
    os.chdir(orig)



"""Flux-local manifest action."""

import pathlib

from flux_local import git_repo


class ManifestAction:
    """Flux-local manifest action."""

    async def run(  # type: ignore[no-untyped-def]
        self,
        path: pathlib.Path,
        **kwargs,  # pylint: disable=unused-argument
    ) -> None:
        """Async Action implementation."""
        manifest = await git_repo.build_manifest(path)
        print(manifest.yaml())



"""Tests for git_repo."""

from pathlib import Path

from flux_local.git_repo import build_manifest

TESTDATA = Path("tests/testdata/cluster")


async def test_build_manifest() -> None:
    """Tests for building the manifest."""

    manifest = await build_manifest(TESTDATA)
    assert len(manifest.clusters) == 1
    assert manifest.clusters[0].name == "flux-system"
    assert manifest.clusters[0].path == "./tests/testdata/cluster/clusters/prod"



"""Flux local test action."""

import asyncio
import logging
from pathlib import Path
from typing import Generator

import nest_asyncio
import pytest

from flux_local import kustomize
from flux_local.manifest import Manifest

_LOGGER = logging.getLogger(__name__)


class KustomizationTest(pytest.Item):
    """Test case for a Kustomization."""

    def runtest(self) -> None:
        """Dispatch the async work and run the test."""
        nest_asyncio.apply()
        asyncio.run(self.async_runtest())

    async def async_runtest(self) -> None:
        """Run the Kustomizations test."""
        await kustomize.build(self.path).objects()


class ManifestCollector(pytest.File):
    """Test collector that generates tests to run from the cluster manifest."""

    def collect(self) -> Generator[pytest.Item, None, None]:
        """Collect tests from the cluster manifest.yaml file."""
        manifest = Manifest.parse_yaml(self.path.read_text(encoding="utf-8"))
        for cluster in manifest.clusters:
            for kustomization in cluster.kustomizations:
                yield KustomizationTest.from_parent(
                    self,
                    name=kustomization.name,
                    path=Path(kustomization.path),
                )


def pytest_collect_file(parent: pytest.Session, file_path: Path) -> pytest.Collector:
    """Build the test of test cases from the cluster manifest."""
    _LOGGER.debug("pytest_collect_file: %s, %s", parent, file_path)
    return ManifestCollector.from_parent(  # type: ignore[no-any-return]
        parent, path=file_path
    )



"""Flux local test action."""

import logging
import pathlib
import tempfile
from pathlib import Path

import nest_asyncio
import pytest

from flux_local import git_repo

_LOGGER = logging.getLogger(__name__)


MANIFEST_FILE = "manifest.yaml"


class TestAction:
    """Flux-local test action."""

    async def run(  # type: ignore[no-untyped-def]
        self,
        path: pathlib.Path,
        **kwargs,  # pylint: disable=unused-argument
    ) -> None:
        """Async Action implementation."""
        with tempfile.TemporaryDirectory() as tmp_dir:
            manifest = await git_repo.build_manifest(path)
            manifest_file = Path(tmp_dir) / MANIFEST_FILE
            manifest_file.write_text(manifest.yaml())
            nest_asyncio.apply()
            pytest.main(
                [
                    "-v",  # Add a line for every test
                    "--no-header",
                    # Load this module only from under pytest
                    "-p",
                    "flux_local.tool.test_kustomization",
                    str(manifest_file),
                ]
            )



"""Tests for the flux-local command line tool."""


from pathlib import Path

from flux_local.command import Command, run

TESTDATA = Path("tests/testdata/cluster/")


async def test_build() -> None:
    """Test flux-local build command."""
    await run(Command(["flux-local", "build", str(TESTDATA)]))


async def test_build_helm() -> None:
    """Test flux-local build command with helm."""
    await run(Command(["flux-local", "build", "--enable-helm", str(TESTDATA)]))


async def test_diff() -> None:
    """Test flux-local diff command."""
    await run(Command(["flux-local", "diff", str(TESTDATA)]))


async def test_manifest() -> None:
    """Test flux-local manifest command."""
    await run(Command(["flux-local", "manifest", str(TESTDATA)]))


async def test_test() -> None:
    """Test flux-local test command."""
    await run(Command(["flux-local", "test", str(TESTDATA)]))



"""Tests for kustomize library."""

from pathlib import Path

import pytest

from flux_local import command, kustomize

TESTDATA_DIR = Path("tests/testdata")


@pytest.mark.parametrize(
    "path",
    [TESTDATA_DIR / "repo", (TESTDATA_DIR / "repo").absolute()],
)
async def test_build(path: Path) -> None:
    """Test a kustomize build command."""
    result = await kustomize.build(path).run()
    assert "Secret" in result
    assert "ConfigMap" in result
    assert result == (TESTDATA_DIR / "repo/all.golden").read_text()


@pytest.mark.parametrize(
    "path",
    [TESTDATA_DIR / "repo", (TESTDATA_DIR / "repo").absolute()],
)
async def test_build_grep(path: Path) -> None:
    """Test a kustomize build and grep command chained."""
    result = await kustomize.build(path).grep("kind=ConfigMap").run()
    assert "Secret" not in result
    assert "ConfigMap" in result
    assert result == (TESTDATA_DIR / "repo/configmap.build.golden").read_text()


@pytest.mark.parametrize(
    "path",
    [TESTDATA_DIR / "repo", (TESTDATA_DIR / "repo").absolute()],
)
async def test_grep(path: Path) -> None:
    """Test a kustomize grep command."""
    result = await kustomize.grep("kind=ConfigMap", path).run()
    assert "Secret" not in result
    assert "ConfigMap" in result
    assert result == (TESTDATA_DIR / "repo/configmap.grep.golden").read_text()


@pytest.mark.parametrize(
    "path",
    [TESTDATA_DIR / "repo", (TESTDATA_DIR / "repo").absolute()],
)
async def test_objects(path: Path) -> None:
    """Test loading yaml documents."""
    cmd = kustomize.build(path).grep("kind=ConfigMap")
    result = await cmd.objects()
    assert len(result) == 1
    assert result[0].get("kind") == "ConfigMap"
    assert result[0].get("apiVersion") == "v1"


@pytest.mark.parametrize(
    "path",
    [TESTDATA_DIR / "repo", (TESTDATA_DIR / "repo").absolute()],
)
async def test_validate_pass(path: Path) -> None:
    """Test applying policies to validate resources."""
    cmd = kustomize.build(path)
    await cmd.validate(TESTDATA_DIR / "policies/pass.yaml")


@pytest.mark.parametrize(
    "path",
    [TESTDATA_DIR / "repo", (TESTDATA_DIR / "repo").absolute()],
)
async def test_validate_fail(path: Path) -> None:
    """Test applying policies to validate resources."""
    cmd = kustomize.build(path)
    with pytest.raises(
        command.CommandException, match="require-test-annotation: validation error"
    ):
        await cmd.validate(TESTDATA_DIR / "policies/fail.yaml")
